import os
import warnings
import pandas as pd
from oemof.tabular.datapackage import building
from datapackage import Package

# TODO when oemof.tabular is updated with inferred metadata PR, use
# from oemof.tabular.config import config
import oemof_tabular_plugins.datapackage.config as config


def map_sequence_profiles_to_resource_name(p, excluded_profiles=("timeindex",)):
    """Look in every resource which is a sequence and map each of its fields to itself

    Within this process the unicity of the field names will be checked, with the exception of the field "timeindex"

    """

    def check_sequences_labels_unicity(labels, new_labels):
        intersect = set(labels).intersection(new_labels)
        if len(intersect) == 1:
            intersect = intersect.pop()
            if not intersect == "timeindex":
                answer = [intersect]
            else:
                answer = []
        else:
            answer = list(intersect)

        if answer:
            warnings.warn(
                f"The labels of the profiles are not unique across all files within 'sequences' folder: '{','.join(intersect)}' used more than once"
            )
        return answer

    sequences = {}
    sequence_labels = []
    duplicated_labels = []
    for r in p.resources:
        if "/sequences/" in r.descriptor["path"]:
            field_labels = [
                f.name for f in r.schema.fields if f.name not in excluded_profiles
            ]
            sequences[r.descriptor["name"]] = field_labels
            duplicated_labels += check_sequences_labels_unicity(
                sequence_labels, field_labels
            )
            sequence_labels += field_labels

    if duplicated_labels:
        # write an error message here
        pass
    # map each profile to its resource name
    sequences_mapping = {
        value: key for (key, values) in sequences.items() for value in values
    }
    return sequences_mapping


def infer_resource_foreign_keys(resource, sequences_profiles_to_resource, busses):
    """Find out the foreign keys within a resource fields

    Look through all field of a resource which are of type 'string' if any of their values are matching a profile header in any of the sequences resources


    Parameters
    ----------
    resource: a :datapackage.Resource: instance
    sequences_profiles_to_resource: the mapping of sequence profile headers to their resource name

    Returns
    -------
    The :datapackage.Resource: instance with updated "foreignKeys" field

    """
    r = resource
    data = pd.DataFrame.from_records(r.read(keyed=True))
    # TODO not sure this should be set here
    r.descriptor["schema"]["primaryKey"] = "name"
    if "foreignKeys" not in r.descriptor["schema"]:
        r.descriptor["schema"]["foreignKeys"] = []

    for field in r.schema.fields:
        if field.type == "string":
            for potential_fk in data[field.name].dropna().unique():

                if potential_fk in sequences_profiles_to_resource:
                    # this is actually a wrong format and should be with a "fields" field under the "reference" fields

                    fk = {
                        "fields": field.name,
                        "reference": {
                            "resource": sequences_profiles_to_resource[potential_fk],
                        },
                    }

                    if fk not in r.descriptor["schema"]["foreignKeys"]:
                        r.descriptor["schema"]["foreignKeys"].append(fk)
                if potential_fk in busses:
                    fk = {
                        "fields": field.name,
                        "reference": {"resource": "bus", "fields": "name"},
                    }
                    if fk not in r.descriptor["schema"]["foreignKeys"]:
                        r.descriptor["schema"]["foreignKeys"].append(fk)

    r.commit()
    return r


def infer_package_foreign_keys(package, typemap=None):
    """Infer the foreign_keys from data/elements and data/sequences and update meta data

    Parameters
    ----------
    package: scenario datapackage
    typemap: facade typemap

    """
    if typemap is None:
        typemap = {}

    p = package
    sequences_profiles_to_resource = map_sequence_profiles_to_resource_name(p)

    bus_data = pd.DataFrame.from_records(p.get_resource("bus").read(keyed=True))

    for r in p.resources:
        # it should be from the folder "elements" and the resource should not be bus.csv
        if (
            "/elements/" in r.descriptor["path"]
            or os.sep + "elements" + os.sep in r.descriptor["path"]
        ) and r.name != "bus":
            r = infer_resource_foreign_keys(
                r, sequences_profiles_to_resource, busses=bus_data.name.to_list()
            )

            if r.name in typemap:
                facade_type = typemap[r.name]
                # test if facade_type has the method 'validate_datapackage'
                if hasattr(facade_type, "validate_datapackage"):
                    # apply the method if it exists
                    facade_type.validate_datapackage(r)

            p.remove_resource(r.name)
            p.add_resource(r.descriptor)


def infer_metadata_from_data(
    package_name="default-name",
    path=None,
    metadata_filename="datapackage.json",
    typemap=None,
):
    """

    Returns
    -------

    """

    # Infer the fields from the package data
    path = os.path.abspath(path)
    p0 = Package(base_path=path)
    p0.infer(os.path.join(path, "**" + os.sep + "*.csv"))
    p0.commit()
    p0.save(os.path.join(path, metadata_filename))

    foreign_keys = {}

    def infer_resource_basic_foreign_keys(resource):
        """insert resource foreign_key into a dict formatted for building.infer_metadata

        Compare the fields of a resource to a list of field names known to be foreign keys. If the field name is within the list, it is used to populate the dict 'foreign_keys'
        """
        for field in resource.schema.fields:
            if field.name in config.SPECIAL_FIELD_NAMES:
                fk_descriptor = config.SPECIAL_FIELD_NAMES[field.name]
                if fk_descriptor in foreign_keys:
                    if resource.name not in foreign_keys[fk_descriptor]:
                        foreign_keys[fk_descriptor].append(resource.name)
                else:
                    foreign_keys[fk_descriptor] = [resource.name]

    for r in p0.resources:
        if os.sep + "elements" + os.sep in r.descriptor["path"]:
            infer_resource_basic_foreign_keys(r)
    # this function saves the metadata of the package in json format
    building.infer_metadata(
        package_name=package_name,
        path=path,
        foreign_keys=foreign_keys,
        metadata_filename=metadata_filename,
    )

    # reload the package from the saved json file
    p = Package(os.path.join(path, metadata_filename))
    infer_package_foreign_keys(p, typemap=typemap)
    p.descriptor["resources"].sort(key=lambda x: (x["path"], x["name"]))
    p.commit()
    p.save(os.path.join(path, metadata_filename))



# TODO: include water (buses, balance, ARID factor)
# TODO: fix documentation
# TODO: compute.py: Filenotfounderror for geometry.json

from dataclasses import field
from typing import Sequence, Union

from oemof.solph._plumbing import sequence
from oemof.solph.buses import Bus
from oemof.solph.components import Converter
from oemof.solph.flows import Flow

from oemof.tabular._facade import Facade, dataclass_facade

from src.oemof_tabular_plugins.wefe.global_specs.crops import crop_dict
from pyomo.environ import ConcreteModel, Var, Objective, SolverFactory, Piecewise, Constraint, maximize
import numpy as np
import pandas as pd
import json
from scipy.interpolate import interp1d


@dataclass_facade
class APVSystem(Converter, Facade):
    r"""APV System unit with 1 input and 2 outputs.

        Parameters
        ----------
        solar_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its input.
        electricity_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its electrical output.
        biomass_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its biomass output.
        water_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its water output.
        capacity: numeric
            The conversion capacity (output side) of the unit.
        marginal_cost: numeric
            Marginal cost for one unit of produced output. Default: 0
        carrier_cost: numeric
            Carrier cost for one unit of used input. Default: 0
        capacity_cost: numeric
            Investment costs per unit of output capacity.
            If capacity is not set, this value will be used for optimizing the
            conversion output capacity.
        expandable: boolean or numeric (binary)
            True, if capacity can be expanded within optimization. Default: False.
        capacity_potential: numeric
            Maximum invest capacity in unit of output capacity.
        capacity_minimum: numeric
            Minimum invest capacity in unit of output capacity.
        input_parameters: dict (optional)
            Set parameters on the input edge of the conversion unit
            (see oemof.solph for more information on possible parameters)
        output_parameters: dict (optional)
            Set parameters on the output edge of the conversion unit
             (see oemof.solph for more information on possible parameters)
        t_air: array-like
            Ambient air temperature
        ghi: array-like
            Global horizontal irradiance
        lat: numeric
            Latitude
        crop_type: string
            Choose crop type from Dictionary in wefe/global_specs/crops.py
        sowing_date: string
            Set sowing date in the format yyyy-mm-dd
        minimal_crop_yield: float
            Minimal crop yield relative to open field, has to be in interval [0, 1]

        SHOULD INCLUDE FUNCTIONS AND EXAMPLE HERE

        """
    solar_bus: Bus

    electricity_bus: Bus

    biomass_bus: Bus

    #water_bus: Bus

    carrier: str

    tech: str

    t_air: Union[float, Sequence[float]]

    ghi: Union[float, Sequence[float]]

    capacity: float = None

    marginal_cost: float = 0

    carrier_cost: float = 0

    capacity_cost: float = None

    expandable: bool = False

    capacity_potential: float = float("+inf")

    capacity_minimum: float = None

    input_parameters: dict = field(default_factory=dict)

    output_parameters: dict = field(default_factory=dict)

    lat: float = 0

    crop_type: str = 'cassava'

    sowing_date: str = '2022-01-01 '

    minimal_crop_yield: float = 0.8

    def build_solph_components(self):
        """ """

        def apv_geometry(instance):
            """ """
            # Location coordinates to obtain geometry parameters
            lat = instance.lat

            # Location-specific, fixed geometry parameters
            # module size from bifacial_radiance 'test_module'
            y = 1.74  # module length (y = N/S)
            x = 1.036  # module width (x = E/W)
            # IBC minimum slope (by means of PV: tilt) for proper rainwater runoff
            min_slope = 0.25 / 12
            min_tilt = np.ceil(np.degrees(np.arctan(min_slope)))
            # tilt should ideally be close to latitude, but allow for rainwater runoff
            tilt = max(round(abs(lat)), min_tilt)
            rad_tilt = np.radians(tilt)
            # minimum solar noon altitude (solar angle at solstice when sun is straight south (lat>0) or north (lat<0)
            min_solar_angle = 90 - round(abs(lat)) - 23.5
            rad_min_solar_angle = np.radians(min_solar_angle)
            # minimum distance to prevent the panels from shading each other
            min_ygap = y * np.sin(rad_tilt) / np.tan(rad_min_solar_angle)
            # define pitch as distance from edge of one module across row up to the edge of the next module
            pitch = round(y * np.cos(rad_tilt) + min_ygap, 2)

            r"""
            Obtain shading and bifaciality factors from global_specs/geometry.json, 
            interpolate for given latitude
            """

            # Load the geometry data
            filepath = 'oemof_tabular_plugins/wefe/global_specs/geometry.json'

            with open(filepath, 'r') as f:
                geometry = json.load(f)

            # Convert lists back to numpy arrays and ensure numerical types
            def convert_from_serializable(obj):
                if isinstance(obj, list):
                    return np.array(obj)
                if isinstance(obj, str) and obj.replace('.', '', 1).isdigit():
                    return float(obj) if '.' in obj else int(obj)
                return obj

            geometry = {float(lat): {k: convert_from_serializable(v)
                                    for k, v in lat_results.items()} for lat, lat_results in geometry.items()}

            def interpolate_values(lat, geometry):
                # Extracting data
                lats = list(geometry.keys())

                # Interpolating fbifacial and fshading for each xgap
                xgaps = geometry[lats[0]]['xgaps'].tolist() \
                    if isinstance(geometry[lats[0]]['xgaps'], np.ndarray) \
                    else geometry[lats[0]]['xgaps']
                fbifacial_interp_funcs = []
                fshading_interp_funcs = []

                for xgap in xgaps:
                    fbifacials = [geometry[lat]['fbifacials'][xgap] for lat in lats]
                    fshadings = [geometry[lat]['fshadings'][xgap] for lat in lats]
                    fbifacial_interp_funcs.append(interp1d(lats, fbifacials, kind='linear'))
                    fshading_interp_funcs.append(interp1d(lats, fshadings, kind='linear'))

                # Calculating new values for the given latitude
                new_fbifacials = [interp_func(lat).item() for interp_func in fbifacial_interp_funcs]
                new_fshadings = [interp_func(lat).item() for interp_func in fshading_interp_funcs]

                return {
                    'xgaps': xgaps,
                    'fbifacials': new_fbifacials,
                    'fshadings': new_fshadings,
                }

            geo_params = interpolate_values(lat, geometry)

            xgaps = geo_params['xgaps']
            fbifacials = geo_params['fbifacials']
            fshadings = geo_params['fshadings']


            r""" 
            Optimize geometry based on analysis results

            Optimization objective is the LER (land equivalent ratio = relative electricity yield + relative crop yield)
            based on the results for bifaciality factors (fbifacials) and shading factors (fshadings) 
            generated by iterating the different gaps between modules (xgaps)

            Area requirement per PV module is a conclusive result.
            """

            # convert the non-linear term for relative electricity yield per area to a piecewise linear function 'farea'
            fareas = [(1 + fbifacials[i]) * x / ((1 + fbifacials[0]) * (x + xgaps[i])) for i in range(len(xgaps))]

            # Define the optimization model
            opti_model = ConcreteModel()

            # Decision variable: xgap
            opti_model.xgap = Var(bounds=(xgaps[0], xgaps[-1]))

            # Set up the piecewise linear functions for shading and area terms
            opti_model.fshading = Var()
            opti_model.fshading_pieces = Piecewise(opti_model.fshading, opti_model.xgap,
                                                   pw_pts=xgaps,
                                                   f_rule=fshadings,
                                                   pw_constr_type='EQ')

            opti_model.farea = Var()
            opti_model.farea_pieces = Piecewise(opti_model.farea, opti_model.xgap,
                                                pw_pts=xgaps,
                                                f_rule=fareas,
                                                pw_constr_type='EQ')

            # Minimum shading constraint
            min_fshading = instance.minimal_crop_yield
            opti_model.min_shading_constraint = Constraint(expr=opti_model.fshading >= min_fshading)

            # Objective: Maximize LER
            def objective_rule(model):
                return model.farea + model.fshading

            opti_model.objective = Objective(rule=objective_rule, sense=maximize)

            # Solver
            solver = SolverFactory('cbc')
            result = solver.solve(opti_model)

            # Output results
            xgap_optimal = opti_model.xgap.value
            fshading_optimal = opti_model.fshading.value
            farea_optimal = opti_model.farea.value
            fbifacial_optimal = ((farea_optimal * (1 + fbifacials[0]) * (xgap_optimal + x)) / x) - 1
            ler_optimal = farea_optimal + fshading_optimal

            # Assign important results as instance attributes
            instance.area_module = x * y * np.cos(rad_tilt)
            instance.area_apv = (xgap_optimal + x) * pitch
            instance.fshading = fshading_optimal
            instance.fbifacial = fbifacial_optimal
            instance.ler = ler_optimal

            return instance

        apv_geometry(self)

        def apv_production(instance):
            r"""
            Calculate biomass and PV efficiency as full-year hourly conversion factor series
            """

            # create date-time-indexed DataFrame with ghi and t_air as columns
            date_time_index = pd.date_range(
                "1/1/2022", periods=8760, freq="H")
            df = pd.DataFrame(index=date_time_index)
            df['ghi'] = instance.ghi
            df['t_air'] = instance.t_air

            # Extract crop specific growth parameters from dictionary
            crop_type = instance.crop_type
            t_opt = crop_dict[crop_type]['t_opt']
            t_base = crop_dict[crop_type]['t_base']
            t_heat = crop_dict[crop_type]['t_max']
            t_extreme = crop_dict[crop_type]['t_ext']
            t_sum = crop_dict[crop_type]['t_sum']
            i50a = crop_dict[crop_type]['i50a']
            i50b = crop_dict[crop_type]['i50b']
            f_solar_max = crop_dict[crop_type]['f_solar_max']
            rue = crop_dict[crop_type]['rue']
            # Convert Radiation Use Efficiency from W/m² to MJ/(m²*h)
            rue *= 3.6e-3
            # Transfer sowing date into Timestamp object
            sowing_date = pd.to_datetime(instance.sowing_date + str(" 01:00:00"))

            # Extract PV parameters
            p_rated = 270  # [Wp]
            rad_ref = 1000  # [W/m²]
            t_ref = 25  # [°C]
            noct = 48  # [°C]

            # Define and apply functions for modelling biomass generation
            def temp(t_air, t_opt, t_base):
                if t_air < t_base:
                    f_temp = 0
                elif t_base <= t_air < t_opt:
                    f_temp = (t_air - t_base) / (t_opt - t_base)
                elif t_air >= t_opt:
                    f_temp = 1
                return f_temp

            df['f_temp'] = df['t_air'].apply(
                lambda t_air: temp(t_air, t_opt, t_base))

            def heat(t_air, t_heat, t_extreme):
                if t_air <= t_heat:
                    f_heat = 1
                elif t_heat < t_air <= t_extreme:
                    f_heat = 1 - (t_air - t_heat) / (t_extreme - t_heat)
                elif t_air > t_extreme:
                    f_heat = 0
                return f_heat

            df['f_heat'] = df['t_air'].apply(
                lambda t_air: heat(t_air, t_heat, t_extreme))

            def development(date, sowing_date, t_air, t_base):
                if date < sowing_date:
                    delta_cum_temp = 0
                elif t_air > t_base:
                    delta_cum_temp = (t_air - t_base) / 24  # SIMPLE crop model has daily temp values, convert to hourly
                return delta_cum_temp

            df['cum_temp'] = df.apply(
                lambda row: development(row.name, sowing_date, row['t_air'], t_base), axis=1).cumsum()

            def solar(cum_temp, t_sum, i50a, i50b, f_solar_max):
                cum_temp_to_reach_f_solar_max = i50a - 100 * np.log(
                    1 / 999)  # derived from f_solar = 0.999f_solar_max
                if cum_temp < 1:
                    f_solar = 0
                elif cum_temp < cum_temp_to_reach_f_solar_max:
                    f_solar = f_solar_max / (1 + np.exp(-0.01 * (cum_temp - i50a)))
                elif cum_temp < t_sum:
                    f_solar = f_solar_max / (1 + np.exp(0.01 * (cum_temp - (t_sum - i50b))))
                else:
                    f_solar = 0
                return f_solar

            df['f_solar'] = df['cum_temp'].apply(
                lambda cum_temp: solar(cum_temp, t_sum, i50a, i50b, f_solar_max))

            df['biomass_efficiency'] = instance.fshading * rue * df['f_solar'] * df['f_temp'] * df['f_heat']

            # Define and apply function for modelling electricity generation
            def power(p_rated, rad, rad_ref, t_air, t_ref, noct):
                f_temp = 1 - 3.7e-3 * (t_air + ((noct - 20) / 800) * rad - t_ref)
                p = p_rated * 1 / rad_ref * f_temp
                return p

            df['pv_efficiency'] = df.apply(
                lambda row: power(p_rated, row['ghi'], rad_ref, row['t_air'], t_ref, noct), axis=1)

            df['pv_efficiency'] *= (1 + instance.fbifacial)

            instance.biomass_efficiency = df['biomass_efficiency']
            instance.pv_efficiency = df['pv_efficiency']

            return instance

        apv_production(self)

        self.conversion_factors.update(
            {
                self.solar_bus: sequence(1),
                self.electricity_bus: sequence(self.pv_efficiency),
                self.biomass_bus: sequence(self.biomass_efficiency)
            }
        )

        self.inputs.update(
            {
                self.solar_bus: Flow(
                    variable_costs=self.carrier_cost, **self.input_parameters
                )
            }
        )

        self.outputs.update(
            {
                self.electricity_bus: Flow(
                    nominal_value=self._nominal_value(),
                    variable_costs=self.marginal_cost,
                    investment=self._investment(),
                    **self.output_parameters,
                ),
                self.biomass_bus: Flow()
            }
        )



import os
import shutil
import pandas as pd
import pytest
from oemof_tabular_plugins.general import pre_processing, calculate_annuity
from unittest.mock import patch
import json


@pytest.mark.parametrize(
    "capex, opex_fix, lifetime, wacc, expected_annuity",
    [(1000000, 50000, 20, 0.05, 130242.59)],
)
def test_calculate_annuity(capex, opex_fix, lifetime, wacc, expected_annuity):
    """
    Tests that the calculate_annuity function works as expected.
    :param capex: CAPEX (currency/MW*) *or the unit you choose to use throughout the model e.g. kW/GW
    :param opex_fix: the fixed OPEX (currency/MW*/year)
    :param lifetime: the lifetime of the component (years)
    :param wacc: the weighted average cost of capital (WACC) applied throughout the model (%)
    :param expected_annuity: the expected annuity value (currency)
    """
    # stores actual result from the calculate_annuity function
    result = calculate_annuity(capex, opex_fix, lifetime, wacc)
    assert result == expected_annuity


class TestPreprocessingCosts:
    """
    This class contains tests for the pre-processing costs function.
    """

    # store the relevant paths
    test_inputs_pre_p = os.path.join("tests", "test_inputs_pre_processing_costs")
    pre_p_dir = os.path.join(
        "tests", "test_inputs_pre_processing_costs", "pre_processing"
    )
    _package_path = os.path.join("data", "elements")
    package_path = _package_path

    def setup_method(self):
        """
        Sets up the testing environment before each test method is executed.
        """
        # removes the 'pre_processing' directory if it exists, ignoring errors
        if os.path.exists(self.pre_p_dir):
            shutil.rmtree(self.pre_p_dir, ignore_errors=True)
        # creates the 'pre_processing' directory
        os.mkdir(self.pre_p_dir)
        # creates the 'data' directory within the 'pre_processing' directory
        os.mkdir(os.path.join(self.pre_p_dir, "data"))
        # creates the 'elements' directory within the 'data' directory
        os.mkdir(os.path.join(self.pre_p_dir, self._package_path))
        # sets the package path for the test environment
        self.package_path = os.path.join(self.pre_p_dir, self._package_path)

    def copy_file_to_package_path(self, f_name):
        """
        Copies the scenario CSV file to the package path
        :param f_name: scenario CSV filename
        """
        # copy scenario csv file and datapackage json file to the package path
        shutil.copy(
            os.path.join(self.test_inputs_pre_p, f_name),
            os.path.join(self.package_path, f_name),
        )

    def teardown_method(self):
        """
        Cleans up the testing environment after each test method is executed.
        """
        # removes the 'pre_processing' directory if it exists, ignoring errors
        if os.path.exists(self.pre_p_dir):
            shutil.rmtree(self.pre_p_dir, ignore_errors=True)
        # resets the package path to its original value
        self.package_path = self._package_path

    def test_annuity_empty_no_cost_params_raises_error(self):
        """
        Tests that a value error is raised when the annuity is left empty and the
        other cost parameters are not included.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_empty_no_cost_params.csv")
        # check if calling the pre_processing function with the given scenario raises a value error
        with pytest.raises(ValueError):
            pre_processing(self.pre_p_dir, wacc=1)

    def test_annuity_defined_no_cost_params_uses_annuity(self):
        """
        Tests that the annuity is directly used when it is defined and the other cost
        parameters are not included.
        """
        # copy scenario csv file to the package path
        f_name = "annuity_defined_no_cost_params.csv"
        self.copy_file_to_package_path(f_name)
        # call pre_processing function
        pre_processing(self.pre_p_dir, wacc=1)
        # check if the actual value matches the expected value
        expected_value = 107265
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        actual_value = df["capacity_cost"].iloc[0]
        assert actual_value == expected_value

    def test_annuity_defined_no_cost_params_logs_message(self, caplog):
        """
        Tests that an info message is logged when the annuity is directly used and the other cost
        parameters are not included.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_defined_no_cost_params.csv")
        # check if the info message is logged when the pre_processing function is called
        pre_processing(self.pre_p_dir, wacc=1)
        assert any(
            record.levelname == "INFO"
            and "The annuity cost is directly used for" in record.message
            for record in caplog.records
        )

    def test_annuity_empty_partial_cost_params_raises_error(self):
        """
        Tests that a value error is raised when the annuity is left empty and not all other cost
        parameters are included.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_empty_partial_cost_params.csv")
        # check if calling the pre_processing function with the given scenario raises a value error
        with pytest.raises(ValueError):
            pre_processing(self.pre_p_dir, wacc=1)

    def test_annuity_defined_partial_cost_params_logs_message(self, caplog):
        """
        Tests that a warning message is logged when the annuity is defined and some but not all other cost
        parameters are included.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_defined_partial_cost_params.csv")
        pre_processing(self.pre_p_dir, wacc=1)
        assert any(
            record.levelname == "WARNING"
            and "directly used but be aware that some cost" in record.message
            for record in caplog.records
        )

    def test_annuity_defined_partial_cost_params_uses_annuity(self):
        """
        Tests that the annuity is directly used when it is defined and some but not all of the other cost
        parameters are included.
        """
        # copy scenario csv file to the package path
        f_name = "annuity_defined_partial_cost_params.csv"
        self.copy_file_to_package_path(f_name)
        # call pre_processing function
        pre_processing(self.pre_p_dir, wacc=1)
        # check if the actual value matches the expected value
        expected_value = 107265
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        actual_value = df["capacity_cost"].iloc[0]
        assert actual_value == expected_value

    def test_annuity_empty_all_cost_params_defined_calculates_annuity(self):
        """
        Tests that the annuity is calculated and added to 'capacity_cost' when the annuity is left
        empty and all other cost parameters are defined.
        """
        # copy scenario csv file to the package path
        f_name = "annuity_empty_all_cost_params_defined.csv"
        self.copy_file_to_package_path(f_name)
        # call the pre_processing function with wacc = 1
        wacc = 1
        pre_processing(self.pre_p_dir, wacc=wacc)
        # check if the actual value matches the expected value
        expected_value = calculate_annuity(
            capex=975000, opex_fix=11625, lifetime=20, wacc=wacc
        )
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        actual_value = df["capacity_cost"].iloc[0]
        assert actual_value == expected_value

    def test_annuity_empty_all_cost_params_defined_logs_message(self, caplog):
        """
        Tests that an info message is logged when a new annuity value is calculated and updated
        from the other cost parameters.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_empty_all_cost_params_defined.csv")
        # check if the info message is logged when the pre_processing function is called
        pre_processing(self.pre_p_dir, wacc=1)
        assert any(
            record.levelname == "INFO"
            and "has been calculated and updated" in record.message
            for record in caplog.records
        )

    def test_annuity_all_cost_params_some_empty_logs_message(self, caplog):
        """
        Tests that warning message is logged when some but not all of the other cost parameters are defined.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_all_cost_params_some_empty.csv")
        # check if the info message is logged when the pre_processing function is called
        pre_processing(self.pre_p_dir, wacc=1)
        assert any(
            record.levelname == "WARNING"
            and "annuity will be directly used but be aware" in record.message
            for record in caplog.records
        )

    def test_annuity_all_cost_params_some_empty_uses_annuity(self):
        """
        Tests that the annuity is directly used when it is defined and some but not all of the other cost
        parameters are defined.
        """
        # copy scenario csv file to the package path
        f_name = "annuity_all_cost_params_some_empty.csv"
        self.copy_file_to_package_path(f_name)
        # call pre_processing function
        pre_processing(self.pre_p_dir, wacc=1)
        # check if the actual value matches the expected value
        expected_value = 107265
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        actual_value = df["capacity_cost"].iloc[0]
        assert actual_value == expected_value

    def test_annuity_defined_all_cost_params_defined_calculates_new_annuity(self):
        """
        Tests that when the annuity is defined and all other cost parameters, the
        new annuity is calculated and replaces the old annuity.
        """
        # copy scenario csv file to the package path
        f_name = "annuity_defined_all_cost_params_defined.csv"
        self.copy_file_to_package_path(f_name)
        wacc = 1
        # call the pre_processing function with wacc = 1
        pre_processing(self.pre_p_dir, wacc=wacc)
        # check if the actual value matches the expected value
        expected_value = calculate_annuity(
            capex=975000, opex_fix=11625, lifetime=20, wacc=1
        )
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        actual_value = df["capacity_cost"].iloc[0]
        assert actual_value == expected_value

    def test_annuity_defined_all_cost_params_defined_logs_message(self, caplog):
        """
        Tests that an info message is logged when the annuity and the other cost
        parameters are defined, so the annuity is replaced by the calculated value
        from the other parameters.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("annuity_defined_all_cost_params_defined.csv")
        wacc = 1
        # check if the info message is logged when the pre_processing function is called
        pre_processing(self.pre_p_dir, wacc=wacc)
        assert any(
            record.levelname == "INFO"
            and "defined annuity has been replaced with" in record.message
            for record in caplog.records
        )

    def test_no_annuity_partial_all_cost_params_empty_raises_error(self):
        """
        Tests that a value error is raised when there is no annuity and not all of the cost
        parameters are defined.
        """
        # copy scenario csv file to the package path
        self.copy_file_to_package_path("no_annuity_partial_all_cost_params_empty.csv")
        # check if calling the pre_processing function with the given scenario raises a value error
        with pytest.raises(ValueError):
            pre_processing(self.pre_p_dir, wacc=1)

    def test_no_annuity_all_cost_params_defined_creates_annuity_param(self):
        """
        Tests that a new annuity parameter ('capacity_cost') is created when it is not included
        and all other cost parameters are defined.
        """
        # copy scenario csv file to the package path
        f_name = "no_annuity_all_cost_params_defined.csv"
        self.copy_file_to_package_path(f_name)
        # call the pre_processing function with wacc=1
        wacc = 1
        pre_processing(self.pre_p_dir, wacc=wacc)
        # check that the 'capacity_cost' parameter is added to the scenario csv file
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        assert "capacity_cost" in df.columns

    def test_no_annuity_all_cost_params_defined_calculates_annuity(self):
        """
        Tests that the annuity is calculated and added when it is not included and all other
        cost parameters are defined.
        """
        # copy scenario csv file to the package path
        f_name = "no_annuity_all_cost_params_defined.csv"
        self.copy_file_to_package_path(f_name)
        # call the pre_processing function with wacc=1
        wacc = 1
        pre_processing(self.pre_p_dir, wacc=wacc)
        # check that the actual value matches the expected value
        expected_value = calculate_annuity(
            capex=975000, opex_fix=11625, lifetime=20, wacc=wacc
        )
        df = pd.read_csv(os.path.join(self.package_path, f_name), delimiter=";")
        actual_value = df["capacity_cost"].iloc[0]
        assert actual_value == expected_value

    def test_no_annuity_all_cost_params_defined_logs_message(self, caplog):
        """
        Tests that an info message is logged when the annuity is not included and all other cost
        parameters are defined.
        """
        # copy scenario csv file to the package path
        f_name = "no_annuity_all_cost_params_defined.csv"
        shutil.copy(
            os.path.join(self.test_inputs_pre_p, f_name),
            os.path.join(self.package_path, f_name),
        )
        # check if the info message is logged when the pre_processing function is called
        pre_processing(self.pre_p_dir, wacc=1)
        assert any(
            record.levelname == "INFO"
            and "has been calculated and updated for" in record.message
            for record in caplog.records
        )

    def test_no_annuity_no_cost_params_logs_message(self, caplog):
        """
        Tests that an info message is logged when neither the annuity or any other cost parameters
        are defined.
        """
        # copy scenario csv file to the package path
        f_name = "no_annuity_no_cost_params.csv"
        shutil.copy(
            os.path.join(self.test_inputs_pre_p, f_name),
            os.path.join(self.package_path, f_name),
        )
        pre_processing(self.pre_p_dir, wacc=1)
        # check if the info message is logged when the pre_processing function is called
        assert any(
            record.levelname == "INFO" and "does not contain" in record.message
            for record in caplog.records
        )


class TestPreprocessingCustomAttributes:
    """
    This class contains tests for the pre-processing custom attributes function.
    """

    # store the relevant paths
    test_inputs_pre_p = os.path.join(
        "tests", "test_inputs_pre_processing_custom_attributes"
    )
    pre_p_dir = os.path.join(
        "tests", "test_inputs_pre_processing_custom_attributes", "pre_processing"
    )
    _package_path = os.path.join("data", "elements")
    package_path = _package_path

    def setup_method(self):
        """
        Sets up the testing environment before each test method is executed.
        """
        # removes the 'pre_processing' directory if it exists, ignoring errors
        if os.path.exists(self.pre_p_dir):
            shutil.rmtree(self.pre_p_dir, ignore_errors=True)
        # creates the 'pre_processing' directory
        os.mkdir(self.pre_p_dir)
        # creates the 'data' directory within the 'pre_processing' directory
        os.mkdir(os.path.join(self.pre_p_dir, "data"))
        # creates the 'elements' directory within the 'data' directory
        os.mkdir(os.path.join(self.pre_p_dir, self._package_path))
        # sets the package path for the test environment
        self.package_path = os.path.join(self.pre_p_dir, self._package_path)

    def copy_files_to_package_path(self, f_name, dp_name):
        """
        Copies the scenario CSV file and the datapackage JSON file to the package path
        :param f_name: scenario CSV filename
        :param dp_name: datapackage JSON filename
        """
        # copy scenario csv file and datapackage json file to the package path
        shutil.copy(
            os.path.join(self.test_inputs_pre_p, f_name),
            os.path.join(self.package_path, f_name),
        )
        shutil.copy(
            os.path.join(self.test_inputs_pre_p, dp_name),
            os.path.join(self.pre_p_dir, "datapackage.json"),
        )

    def teardown_method(self):
        """
        Cleans up the testing environment after each test method is executed.
        """
        # removes the 'pre_processing' directory if it exists, ignoring errors
        if os.path.exists(self.pre_p_dir):
            shutil.rmtree(self.pre_p_dir, ignore_errors=True)
        # resets the package path to its original value
        self.package_path = self._package_path

    def test_output_params_added_to_csv_with_cust_attr_in_csv_and_list(self):
        """
        Tests that the output parameters column has been added to the csv file if the csv
        file contains custom attributes and they have been defined as list by user.
        """
        # copy scenario csv file and datapackage json file to the package path
        f_name = "cust_attr.csv"
        self.copy_files_to_package_path(f_name, "dp_no_output_params.json")
        # call the pre_processing function with wacc = 1 and custom_attributes list defined
        wacc = 1
        pre_processing(
            self.pre_p_dir,
            wacc=wacc,
            custom_attributes=[
                "emission_factor",
                "renewable_factor",
                "land_requirement",
            ],
        )
        # read the updated csv file
        updated_df = pd.read_csv(os.path.join(self.package_path, f_name), sep=";")
        # assert that 'output_parameters' column is in the updated dataframe
        assert "output_parameters" in updated_df.columns, (
            "'output_parameters' column is not present " "in the updated dataframe"
        )

    def test_output_params_not_added_to_csv_with_cust_attr_in_csv_and_not_list(self):
        """
        Tests that the output parameters column has not been added to the csv file if the csv
        file contains custom attributes and they have not been defined as list by user.
        """
        # copy scenario csv file and datapackage json file to the package path
        f_name = "cust_attr.csv"
        self.copy_files_to_package_path(f_name, "dp_no_output_params.json")
        # read the original dataframe before pre-processing
        original_df = pd.read_csv(os.path.join(self.package_path, f_name), sep=";")
        # call the pre_processing function with wacc = 1 and custom_attributes = none (default)
        wacc = 1
        pre_processing(self.pre_p_dir, wacc=wacc)
        # read the updated csv file after pre-processing
        updated_df = pd.read_csv(os.path.join(self.package_path, f_name), sep=";")
        # assert that 'output_parameters' column is not in the updated dataframe if it wasn't present initially
        if "output_parameters" not in original_df.columns:
            assert (
                "output_parameters" not in updated_df.columns
            ), "'output_parameters' has been added to the updated dataframe when it should not be"

    def test_output_params_not_added_to_twice_when_already_exists(self):
        """
        Tests that the output parameters field has not been added again to the datapackage.json file
        when it already exists
        """
        # copy scenario csv file and datapackage json file to the package path
        self.copy_files_to_package_path("cust_attr.csv", "dp_output_params.json")
        # call the pre_processing function with wacc = 1 and custom_attributes list defined
        wacc = 1
        pre_processing(
            self.pre_p_dir,
            wacc=wacc,
            custom_attributes=[
                "emission_factor",
                "renewable_factor",
                "land_requirement",
            ],
        )
        # read the datapackage.json file after pre-processing
        with open(os.path.join(self.pre_p_dir, "datapackage.json"), "r") as f:
            updated_datapackage = json.load(f)
        # get the resource from the datapackage.json file
        resource = updated_datapackage.get("resources", [None])[0]
        # check if the resource was found
        assert (
            resource is not None
        ), "No resource found in the updated datapackage.json file"
        # find the appropriate field within the resource's schema
        fields = resource.get("schema", {}).get("fields", [])
        output_parameters_count = sum(
            1 for field in fields if field.get("name") == "output_parameters"
        )
        # assert that the 'output_parameters' field exists only once
        assert (
            output_parameters_count == 1
        ), "The 'output_parameters' field exists more than once in the schema"



from oemof.tabular.config import config

SPECIAL_FIELD_NAMES = {}
for fk, descriptor in config.FOREIGN_KEY_DESCRIPTORS.items():
    for el in descriptor:
        SPECIAL_FIELD_NAMES[el["fields"]] = fk



from pyomo import environ as po


def renewable_share_minimum(model, flows=None, limit=None):
    keyword = "renewable_factor"

    if flows is None:
        flows = {}
        for i, o in model.flows:
            if hasattr(model.flows[i, o], keyword):
                flows[(i, o)] = model.flows[i, o]
    else:
        for i, o in flows:
            if not hasattr(flows[i, o], keyword):
                raise AttributeError(
                    f"Flow with source: {i.label} and target: {o.label} has no attribute {keyword}."
                )
    # this structure only works if you don't use multi-period
    total_generation = sum(
        model.flow[i, o, 0, t] for (i, o) in flows for t in model.TIMESTEPS
    )
    renewable_generation = sum(
        model.flow[i, o, 0, t] * getattr(flows[i, o], keyword)
        for (i, o) in flows
        for t in model.TIMESTEPS
    )
    constraint_name = "renewable_share_minimum_constraint"
    model.add_component(
        constraint_name,
        po.Constraint(expr=(renewable_generation >= float(limit) * total_generation)),
    )

    return model



# TODO: assign water_bus output independent of solar_bus (absolute output instead of conversion factor)
# TODO: install bifacial radiance (rather get rid of it anyway)
# TODO: fix documentation

from dataclasses import field
from typing import Sequence, Union

from oemof.solph._plumbing import sequence
from oemof.solph.buses import Bus
from oemof.solph.components import Converter
from oemof.solph.flows import Flow

from oemof.tabular._facade import Facade, dataclass_facade

from src.oemof_tabular_plugins.wefe.global_specs.crops import crop_dict
from bifacial_radiance import *
from pyomo.environ import ConcreteModel, Var, Objective, SolverFactory, Piecewise, Constraint, maximize
import numpy as np
import pandas as pd
import os
from pathlib import Path


@dataclass_facade
class APVSystem(Converter, Facade):
    r"""APV System unit with 1 input and 2 outputs.

        Parameters
        ----------
        solar_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its input.
        electricity_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its electrical output.
        biomass_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its biomass output.
        water_bus: oemof.solph.Bus
            An oemof bus instance where the APV system unit is connected to with
            its water output.
        capacity: numeric
            The conversion capacity (output side) of the unit.
        marginal_cost: numeric
            Marginal cost for one unit of produced output. Default: 0
        carrier_cost: numeric
            Carrier cost for one unit of used input. Default: 0
        capacity_cost: numeric
            Investment costs per unit of output capacity.
            If capacity is not set, this value will be used for optimizing the
            conversion output capacity.
        expandable: boolean or numeric (binary)
            True, if capacity can be expanded within optimization. Default: False.
        capacity_potential: numeric
            Maximum invest capacity in unit of output capacity.
        capacity_minimum: numeric
            Minimum invest capacity in unit of output capacity.
        input_parameters: dict (optional)
            Set parameters on the input edge of the conversion unit
            (see oemof.solph for more information on possible parameters)
        output_parameters: dict (optional)
            Set parameters on the output edge of the conversion unit
             (see oemof.solph for more information on possible parameters)
        t_air: array-like
            Ambient air temperature
        ghi: array-like
            Global horizontal irradiance
        lat: numeric
            Latitude
        lon: numeric
            Longitude
        reference_year: numeric
            Reference year for radiance simulation, if None: average over multiple years
        simulation_name: string
            Custom name for radiance simulation, if None: generic name based on lat and lon
        crop_type: string
            Choose crop type from Dictionary in ...
        sowing_date: string
            Set sowing date in the format yyyy-mm-dd
        minimal_crop_yield: float
            Minimal crop yield relative to open field, has to be in interval [0, 1]

        SHOULD INCLUDE FUNCTIONS AND EXAMPLE HERE

        """
    solar_bus: Bus

    electricity_bus: Bus

    biomass_bus: Bus

    water_bus: Bus

    carrier: str

    tech: str

    t_air: Union[float, Sequence[float]]

    ghi: Union[float, Sequence[float]]

    capacity: float = None

    marginal_cost: float = 0

    carrier_cost: float = 0

    capacity_cost: float = None

    expandable: bool = False

    capacity_potential: float = float("+inf")

    capacity_minimum: float = None

    input_parameters: dict = field(default_factory=dict)

    output_parameters: dict = field(default_factory=dict)

    lat: float = 0

    lon: float = 0

    reference_year: int = None

    simulation_name: str = None

    crop_type: str = 'cassava'

    sowing_date: str = '2022-01-01 '

    minimal_crop_yield: float = 0.8

    def build_solph_components(self):
        """ """

        def apv_geometry(instance):
            # TODO: sensitivity, steps and step_width as variable arguments?

            r"""
             ***** Part 1: Define general geometry parameters and initiate radiance model *****
             """
            # Location coordinates and reference year for radiance simulation
            lat = instance.lat
            lon = instance.lon
            year = instance.reference_year

            # Simulation name
            if instance.simulation_name != None:
                sim_name = instance.simulation_name
            else:
                sim_name = "lat" + str(lat) + "lon" + str(lon)

            # Create directory within project for radiation analysis
            folder = Path().resolve() / 'bifacial_radiance' / 'TEMP' / sim_name
            print(f"Your simulation will be stored in {folder}")

            if not os.path.exists(folder):
                os.makedirs(folder)

            # Create Radiance object
            rad_model = RadianceObj(name=sim_name, path=str(folder))

            # Get weather data
            epwfile = rad_model.getEPW(lat=lat, lon=lon)
            metdata = rad_model.readWeatherFile(weatherFile=epwfile, coerce_year=year)

            # Ground albedo and sky
            albedo = 0.23  # FAO56
            rad_model.setGround(material=albedo)  # must be run before for .gencumsky or .gendaylit
            rad_model.genCumSky()

            # Module and scene
            moduletype = 'test_module'
            # all parameters derived from module.json, x=1.036 and y=1.74, portrait mode, numpanels=1
            y = 1.74  # module length (y = N/S)
            x = 1.036  # module width (x = E/W)

            # panel facing south if lat >= 0, else facing north
            azimuth = 180 if lat >= 0 else 0
            # IBC minimum slope (by means of PV: tilt) for proper rainwater runoff
            min_slope = 0.25 / 12
            min_tilt = np.ceil(np.degrees(np.atan(min_slope)))
            # tilt should ideally be close to latitude, but allow for rainwater runoff
            tilt = max(round(abs(lat)), min_tilt)
            rad_tilt = np.radians(tilt)
            # minimum solar noon altitude (solar angle at solstice when sun is straight south (lat>0) or north (lat<0)
            min_solar_angle = 90 - round(abs(lat)) - 23.5
            rad_min_solar_angle = np.radians(min_solar_angle)
            # minimum distance to prevent the panels from shading each other
            min_ygap = round(y * np.sin(rad_tilt) / np.tan(rad_min_solar_angle), 2)
            # define pitch as distance from edge of one module across row up to the edge of the next module
            pitch = y * np.cos(rad_tilt) + min_ygap
            # clearance height fixed at 3 m (no significant change afterwards; see paper Sánchez, Meza, Dittmann; own iteration)
            clearance_height = 3
            # numer of Rows and Modules per row fixed at 5 each, no significant change afterwards; own iteration results)
            n = 15
            # analysis sensitivity (number of sensors in x and y direction on panel and ground)
            sensors = 10
            # iteration steps and width
            steps = 5
            step_width = 1

            # Fill SceneDict with constant geometry parameters
            sceneDict = {'tilt': tilt,
                         'pitch': pitch,
                         'clearance_height': clearance_height,
                         'azimuth': azimuth,
                         'nMods': n,
                         'nRows': n
                         }

            r"""
            Part 2: Analyze radiance model for varying gaps between PV modules
            """

            # Create lists for analysis results:
            xgaps = []
            fbifacials = []
            fshadings = []

            # Iterate through xgaps and calculate shading and bifaciality factors
            for step in range(steps):
                xgap = step * step_width
                # Create module based on xgap; create scene based on module
                module = rad_model.makeModule(name=moduletype, x=x, y=y, xgap=xgap)
                scene = rad_model.makeScene(module=module, sceneDict=sceneDict)

                # Combine everything in .oct file for analysis
                oct = rad_model.makeOct()

                # Create analysis object
                analysis = AnalysisObj(octfile=oct, name=rad_model.basename)

                # Create scans on the panel (automatic sensor positioning)
                frontscan, backscan = analysis.moduleAnalysis(scene=scene, sensorsy=sensors, sensorsx=sensors)

                # Copy panel scans and adjust for the ground
                groundfrontscan = frontscan.copy()

                # setting height to 12 cm from the ground (FAO56), constant height
                groundfrontscan['zstart'] = 0.12
                groundfrontscan['zinc'] = 0
                # keep first x, increase x spacing, so it covers 1 module and gap to next module
                groundfrontscan['xinc'] = (x + xgap) / (sensors - 1)
                # keep first y, increase y spacing, so it covers 1 pitch
                groundfrontscan['yinc'] = pitch / (sensors - 1)

                groundbackscan = groundfrontscan.copy()

                # Panel analysis
                analysis.analysis(octfile=oct, name=rad_model.basename + f"_panelscan_{step}",
                                  frontscan=frontscan, backscan=backscan)
                fbifacial = np.mean(analysis.backRatio)
                # fshadingpv = np.mean(analysis.Wm2Front) / metdata.ghi.sum()

                # Ground analysis
                analysis.analysis(octfile=oct, name=rad_model.basename + f"_groundscan_{step}",
                                  frontscan=groundfrontscan, backscan=groundbackscan)
                fshading = np.mean(analysis.Wm2Front) / metdata.ghi.sum()

                # Store results in lists
                xgaps.append(xgap)
                fbifacials.append(fbifacial)
                fshadings.append(fshading)

            r""" 
            Part 3: Optimize geometry based on analysis results

            Optimization objective is the LER (land equivalent ratio = relative electricity yield + relative crop yield)
            based on the results for bifaciality factors (fbifacials) and shading factors (fshadings) 
            generated by iterating through the different gaps between modules (xgaps)

            Area requirement per PV module is a conclusive result.
            """

            # convert the non-linear term for relative electricity yield per area to a piecewise linear function 'farea'
            fareas = [(1 + fbifacials[i]) * x / ((1 + fbifacials[0]) * (x + xgaps[i])) for i in range(len(xgaps))]

            # Define the optimization model
            opti_model = ConcreteModel()

            # Decision variable: xgap
            opti_model.xgap = Var(bounds=(xgaps[0], xgaps[-1]))

            # Set up the piecewise linear functions for shading and area terms
            opti_model.fshading = Var()
            opti_model.fshading_pieces = Piecewise(opti_model.fshading, opti_model.xgap,
                                                   pw_pts=xgaps,
                                                   f_rule=fshadings,
                                                   pw_constr_type='EQ')

            opti_model.farea = Var()
            opti_model.farea_pieces = Piecewise(opti_model.farea, opti_model.xgap,
                                                pw_pts=xgaps,
                                                f_rule=fareas,
                                                pw_constr_type='EQ')

            # Minimum shading constraint
            min_fshading = instance.minimal_crop_yield
            opti_model.min_shading_constraint = Constraint(expr=opti_model.fshading >= min_fshading)

            # Objective: Maximize LER
            def objective_rule(model):
                return model.farea + model.fshading

            opti_model.objective = Objective(rule=objective_rule, sense=maximize)

            # Solver
            solver = SolverFactory('cbc')
            result = solver.solve(opti_model)

            # Output results
            xgap_optimal = opti_model.xgap.value
            fshading_optimal = opti_model.fshading.value
            farea_optimal = opti_model.farea.value
            fbifacial_optimal = ((farea_optimal * (1 + fbifacials[0]) * (xgap_optimal + x)) / x) - 1
            ler_optimal = farea_optimal + fshading_optimal

            # Assign important results as instance attributes
            instance.area_module = x * y * np.cos(rad_tilt)
            instance.area_apv = (xgap_optimal + x) * pitch
            instance.f_shading = fshading_optimal
            instance.f_bifacial = fbifacial_optimal
            instance.ler = ler_optimal

            return instance

        apv_geometry(self)

        def apv_production(instance):
            r"""

            :param sowing_date:
            :param crop_type:
            :param xgaps:
            :param fshadings:
            :param fbifacials:
            :param rad:
            :param t_air:
            :return:
            """

            # create date-time-indexed DataFrame with ghi and t_air as columns
            date_time_index = pd.date_range(
                "1/1/2022", periods=8760, freq="H")
            df = pd.DataFrame(index=date_time_index)
            df['ghi'] = instance.ghi
            df['t_air'] = instance.t_air

            # Extract crop specific growth parameters from dictionary
            crop_type = instance.crop_type
            t_opt = crop_dict[crop_type]['t_opt']
            t_base = crop_dict[crop_type]['t_base']
            t_heat = crop_dict[crop_type]['t_max']
            t_extreme = crop_dict[crop_type]['t_ext']
            t_sum = crop_dict[crop_type]['t_sum']
            i50a = crop_dict[crop_type]['i50a']
            i50b = crop_dict[crop_type]['i50b']
            f_solar_max = crop_dict[crop_type]['f_solar_max']
            rue = crop_dict[crop_type]['rue']
            # Convert Radiation Use Efficiency from W/m² to MJ/(m²*h)
            rue *= 3.6e-3
            # Transfer sowing date into Timestamp object
            sowing_date = pd.to_datetime(instance.sowing_date + str(" 01:00:00"))

            # Extract PV parameters
            p_rated = 270  # [Wp]
            rad_ref = 1000  # [W/m²]
            t_ref = 25  # [°C]
            noct = 48  # [°C]

            # Define and apply functions for modelling biomass generation
            def temp(t_air, t_opt, t_base):
                if t_air < t_base:
                    f_temp = 0
                elif t_base <= t_air < t_opt:
                    f_temp = (t_air - t_base) / (t_opt - t_base)
                elif t_air >= t_opt:
                    f_temp = 1
                return f_temp

            df['f_temp'] = df['t_air'].apply(
                lambda t_air: temp(t_air, t_opt, t_base))

            def heat(t_air, t_heat, t_extreme):
                if t_air <= t_heat:
                    f_heat = 1
                elif t_heat < t_air <= t_extreme:
                    f_heat = 1 - (t_air - t_heat) / (t_extreme - t_heat)
                elif t_air > t_extreme:
                    f_heat = 0
                return f_heat

            df['f_heat'] = df['t_air'].apply(
                lambda t_air: heat(t_air, t_heat, t_extreme))

            def development(date, sowing_date, t_air, t_base):
                if date < sowing_date:
                    delta_cum_temp = 0
                elif t_air > t_base:
                    delta_cum_temp = (t_air - t_base) / 24  # SIMPLE crop model has daily temp values, convert to hourly
                return delta_cum_temp

            df['cum_temp'] = df.apply(
                lambda row: development(row.name, sowing_date, row['t_air'], t_base), axis=1).cumsum()

            def solar(cum_temp, t_sum, i50a, i50b, f_solar_max):
                cum_temp_to_reach_f_solar_max = i50a - 100 * np.log(
                    1 / 999)  # derived from f_solar = 0.999f_solar_max
                if cum_temp < 1:
                    f_solar = 0
                elif cum_temp < cum_temp_to_reach_f_solar_max:
                    f_solar = f_solar_max / (1 + np.exp(-0.01 * (cum_temp - i50a)))
                elif cum_temp < t_sum:
                    f_solar = f_solar_max / (1 + np.exp(0.01 * (cum_temp - (t_sum - i50b))))
                else:
                    f_solar = 0
                return f_solar

            df['f_solar'] = df['cum_temp'].apply(
                lambda cum_temp: solar(cum_temp, t_sum, i50a, i50b, f_solar_max))

            df['biomass_efficiency'] = rue * df['f_solar'] * df['f_temp'] * df['f_heat']

            # Define and apply function for modelling electricity generation
            def power(p_rated, rad, rad_ref, t_air, t_ref, noct):
                f_temp = 1 - 3.7e-3 * (t_air + ((noct - 20) / 800) * rad - t_ref)
                p = p_rated * 1 / rad_ref * f_temp
                return p

            df['pv_efficiency'] = df.apply(
                lambda row: power(p_rated, row['ghi'], rad_ref, row['t_air'], t_ref, noct), axis=1)

            instance.biomass_efficiency = df['biomass_efficiency']
            instance.pv_efficiency = df['pv_efficiency']
            # instance.water

        apv_production(self)

        self.conversion_factors.update(
            {
                self.solar_bus: sequence(1),
                self.electricity_bus: sequence(self.pv_efficiency),
                self.biomass_bus: sequence(self.biomass_efficiency)
            }
        )

        self.inputs.update(
            {
                self.from_bus: Flow(
                    variable_costs=self.carrier_cost, **self.input_parameters
                )
            }
        )

        self.outputs.update(
            {
                self.electricity_bus: Flow(
                    nominal_value=self._nominal_value(),
                    variable_costs=self.marginal_cost,
                    investment=self._investment(),
                    **self.output_parameters,
                ),
                self.biomass_bus: Flow()
            }
        )



import logging

import pandas as pd
from datapackage import Package
import oemof.solph as solph
import numpy as np

# ToDo: check to see if the storage optimized input/output (invest_out) and
#  optimized capacity (invest) are saved correctly
# ToDo: is another raw output from the results is investment costs? or does this have to be calculated?
RAW_OUTPUTS = ["investments"]
PROCESSED_RAW_OUTPUTS = ["flow_min", "flow_max", "aggregated_flow"]
RAW_INPUTS = [
    "marginal_cost",
    "carrier_cost",
    "capacity_cost",
    "storage_capacity_cost",
    "capacity",
    "capacity_potential",
    "capacity_minimum",
    "expandable",
    "storage_capacity",
    "storage_capacity_potential",
    "efficiency",
    "capex",
    "opex_fix",
    "lifetime",
    "renewable_factor",
    "emission_factor",
    "land_requirement_factor",
    "water_footprint_factor",
]


# Functions for results per component
def compute_capacity_total(results_df):
    # ToDo: how it is here is that now total capacity is considering the storage capacity (in MWh) for storage.
    #  Check how the storage capacities should be displayed in the results to make it not confusing for the user. Maybe
    #  the storage components need two total capacity results (one for power and one for energy)?
    """Calculates total capacity by adding existing capacity (capacity) to optimized capacity (investments)"""
    if "storage" in results_df.name:
        return results_df.storage_capacity + results_df.investments
    else:
        return results_df.capacity + results_df.investments


def compute_annuity_total(results_df):
    """Calculates total annuity by multiplying the annuity by the optimized capacity"""
    # ToDo: now storage_capacity_cost is used for the annuity if the component is storage.
    #  Check that this is correctly applied for storage components or if two different costs should
    #  be calculated (one for power and one for energy)
    if "storage" in results_df.name:
        return results_df.storage_capacity_cost * results_df.investments
    else:
        return results_df.capacity_cost * results_df.investments


def compute_upfront_investment_costs(results_df):
    # ToDo: check for storage if investments is based on correct parameter
    """Calculates investment costs by multiplying capex with optimized capacity (investments)"""
    if "capex" not in results_df.index:
        return None
    else:
        return results_df.capex * results_df.investments


def compute_opex_fix_costs(results_df):
    """Calculates yearly opex costs by multiplying opex with optimized capacity (investments)"""
    if "opex_fix" not in results_df.index:
        return None
    else:
        return results_df.opex_fix * results_df.investments


def compute_variable_costs(results_df):
    """Calculates variable costs by multiplying the marginal cost by the aggregated flow if the direction is out,
    and by the carrier cost if the direction is in. The total marginal costs for each asset correspond to the sum
    of the marginal costs for the in- and output flows"""
    if results_df.name[1] == "out":
        if "marginal_cost" not in results_df.index:
            return None
        return results_df.marginal_cost * results_df.aggregated_flow
    elif results_df.name[1] == "in":
        if "carrier_cost" not in results_df.index:
            return None
        return results_df.carrier_cost * results_df.aggregated_flow


def compute_renewable_generation(results_df):
    """Calculates renewable generation by multiplying aggregated flow by renewable factor"""
    if "renewable_factor" not in results_df.index:
        return None
    else:
        return results_df.aggregated_flow * results_df.renewable_factor


def compute_co2_emissions(results_df):
    """Calculates CO2 emissions by multiplying aggregated flow by emission factor"""
    if "emission_factor" not in results_df.index:
        return None
    else:
        return results_df.aggregated_flow * results_df.emission_factor


def compute_land_requirement_additional(results_df):
    """Calculates land requirement needed for optimized capacities"""
    if "land_requirement_factor" not in results_df.index:
        return None
    else:
        return results_df.investments * results_df.land_requirement_factor


def compute_land_requirement_total(results_df):
    """Calculates land requirement needed for total capacities"""
    if "land_requirement_factor" not in results_df.index:
        return None
    else:
        return results_df.capacity_total * results_df.land_requirement_factor


def compute_water_footprint(results_df):
    """Calculates water footprint by multiplying aggregated flow by water footprint factor"""
    if "water_footprint_factor" not in results_df.index:
        return None
    else:
        return results_df.aggregated_flow * results_df.water_footprint_factor


# Functions for whole system results
def compute_system_annuity_total(results_df):
    """Calculates system total annuity by summing the total annuity for each component"""
    # ToDo: this method looks through each component and if it is mentioned twice, the annuity only
    #  gets considered once e.g. for storage, except for a MIMO because the costs only get considered once already.
    #  This is a quick fix and I didn't have time to figure out how this should be done in the cleanest way
    seen_components = set()
    annuity_total = 0
    for index, row in results_df.iterrows():
        component_name = index[2]
        # check if the component has been included before
        if component_name not in seen_components:
            annuity_value = row["annuity_total"]
            if pd.isna(annuity_value):
                annuity_value = 0
            annuity_total += annuity_value
            # this is a quick fix to not include the MIMO converter because the asset type is 'nan'
            # this should definitely be changed once implemented properly
            if not pd.isna(index[4]):
                seen_components.add(component_name)
    return annuity_total


def compute_system_variable_costs_total(results_df):
    """Calculates the total variable costs by summing the variable costs for each component flow"""
    # This function has not been implemented the same as above because here we want to consider the variable
    # costs attached to each flow instead of each component
    variable_costs_total = results_df["variable_costs_total"].sum()
    return variable_costs_total


def compute_system_cost_total(results_df):
    """Calculates the total system cost by summing the total annuity and total variable costs
    for each component"""
    # ToDo: quick fix - I didn't have time but rather than repeating the functions above, it would be good to calculate
    #  this from the kpis dataframe instead of results_df (then can use the annuity_total and variable_costs_total
    #  directly. To do this though, the apply_kpi_calculations function has to be adapted
    seen_components = set()
    annuity_total = 0
    for index, row in results_df.iterrows():
        component_name = index[2]
        # check if the component has been included before
        if component_name not in seen_components:
            annuity_value = row["annuity_total"]
            if pd.isna(annuity_value):
                annuity_value = 0
            annuity_total += annuity_value
            # this is a quick fix to not include the MIMO converter because the asset type is 'nan'
            # this should definitely be changed once implemented properly
            if not pd.isna(index[4]):
                seen_components.add(component_name)
    variable_costs_total = results_df["variable_costs_total"].sum()
    system_cost_total = annuity_total + variable_costs_total
    return system_cost_total


def compute_system_upfront_investments_total(results_df):
    """Calculates the total upfront investments by summing the upfront investments for each component"""
    # ToDo: this method looks through each component and if it is mentioned twice, the annuity only
    #  gets considered once e.g. for storage, except for a MIMO because the costs only get considered once already.
    #  This is a quick fix and I didn't have time to figure out how this should be done in the cleanest way
    seen_components = set()
    upfront_investments_total = 0
    for index, row in results_df.iterrows():
        component_name = index[2]
        # check if the component has been included before
        if component_name not in seen_components:
            upfront_investment = row["upfront_investment_costs"]
            if pd.isna(upfront_investment):
                upfront_investment = 0
            upfront_investments_total += upfront_investment
            # this is a quick fix to not include the MIMO converter because the asset type is 'nan'
            # this should definitely be changed once implemented properly
            if not pd.isna(index[4]):
                seen_components.add(component_name)
    return upfront_investments_total


def compute_system_co2_emissions_total(results_df):
    """Calculates the total CO2 emissions by summing up the CO2 emissions on each component flow"""
    # ToDo: so far these are simply summed for each flow, but should check this is correct in every case
    emissions_total = results_df["co2_emissions"].sum()
    return emissions_total


def compute_system_land_requirement_additional(results_df):
    """Calculates the additional land requirement from optimized capacities by summing the additional
    land requirement for each component"""
    # ToDo: this method looks through each component and if it is mentioned twice, the annuity only
    #  gets considered once e.g. for storage, except for a MIMO because the costs only get considered once already.
    #  This is a quick fix and I didn't have time to figure out how this should be done in the cleanest way
    seen_components = set()
    add_land_requirement_total = 0
    for index, row in results_df.iterrows():
        component_name = index[2]
        # check if the component has been included before
        if component_name not in seen_components:
            add_land_requirement = row["land_requirement_additional"]
            if pd.isna(add_land_requirement):
                add_land_requirement = 0
            add_land_requirement_total += add_land_requirement
            # this is a quick fix to not include the MIMO converter because the asset type is 'nan'
            # this should definitely be changed once implemented properly
            if not pd.isna(index[4]):
                seen_components.add(component_name)
    return add_land_requirement_total


def compute_system_land_requirement_total(results_df):
    """Calculates the total land requirement by summing the total land requirement for each component"""
    # ToDo: this method looks through each component and if it is mentioned twice, the annuity only
    #  gets considered once e.g. for storage, except for a MIMO because the costs only get considered once already.
    #  This is a quick fix and I didn't have time to figure out how this should be done in the cleanest way
    seen_components = set()
    land_requirement_total = 0
    for index, row in results_df.iterrows():
        component_name = index[2]
        # check if the component has been included before
        if component_name not in seen_components:
            land_requirement = row["land_requirement_total"]
            if pd.isna(land_requirement):
                land_requirement = 0
            land_requirement_total += land_requirement
            # this is a quick fix to not include the MIMO converter because the asset type is 'nan'
            # this should definitely be changed once implemented properly
            if not pd.isna(index[4]):
                seen_components.add(component_name)
    return land_requirement_total


def compute_water_footprint_total(results_df):
    """Calculates the total water footprint by summing the total water footprint for each component"""
    # ToDo: so far these are simply summed for each flow, but should check this is correct in every case
    water_footprint_total = results_df["water_footprint"].sum()
    return water_footprint_total


def compute_specific_system_cost(results_df):
    """Calculates the total upfront investments by summing the upfront investments for each component"""
    # ToDo: will need to be adapted when non-energetic loads are included - for now only electricity is
    #  considered but this is not correct
    # ToDo: need to decide how this should be calculated for energy systems with multiple carriers
    #  (both energetic and non-energetic)
    total_load = 0
    total_system_cost = (
        results_df["total_annuity"].sum() + results_df["total_variable_costs"].sum()
    )
    for index, row in results_df.iterrows():
        # This is a quick fix to not include water - need to talk to Julian about how other demands should
        # be considered
        if index[4] == "load" and index[3] == "electricity":
            total_load += row.get("aggregated_flow", 0)
    specific_system_cost = total_system_cost / total_load
    return specific_system_cost


def compute_renewable_share(results_df):
    """Calculates the renewable share based on the renewable generation of each flow and the
    total aggregated flow of any component where the renewable factor is set (should be only set on sources)
    """
    # ToDo: this might need to be reconsidered when the renewable share is set on a non-source component
    #  e.g. if the PV panel is a transformer component and the renewable share is on the output. It might still
    #  work but definitely needs to be checked
    renewable_generation_total = 0
    generation_total = 0
    for index, row in results_df.iterrows():
        if not pd.isna(row["renewable_factor"]):
            generation = row["aggregated_flow"]
            renewable_generation = row["aggregated_flow"] * row["renewable_factor"]
            generation_total += generation
            renewable_generation_total += renewable_generation
    if generation_total == 0:
        renewable_share = 0
    else:
        renewable_share = renewable_generation_total / generation_total
    return renewable_share


def _check_arguments(df, column_names, col_name):
    """Check that all required argument are present in the DataFrame columns"""
    for arg in column_names:
        if arg not in df.columns:
            raise AttributeError(
                f"The column {arg} is not present within the results DataFrame and is required to compute '{col_name}', listed in the calculations to be executed"
            )


# TODO turn the dict into a class simular to the one of Calculation of oemof.tabular
CALCULATED_OUTPUTS = [
    {
        "column_name": "capacity_total",
        "operation": compute_capacity_total,
        "description": "The total capacity is calculated by adding the optimized capacity (investments) "
        "to the existing capacity (capacity)",
        "argument_names": ["investments", "capacity"],
    },
    {
        "column_name": "annuity_total",
        "operation": compute_annuity_total,
        "description": "Total annuity is calculated by multiplying the optimized capacity "
        "by the capacity cost (annuity considering CAPEX, OPEX and WACC)",
        "argument_names": ["investments", "capacity_cost"],
    },
    {
        "column_name": "upfront_investment_costs",
        "operation": compute_upfront_investment_costs,
        "description": "Upfront investment costs are calculated by multiplying the optimized capacity "
        "by the CAPEX",
        "argument_names": ["investments", "capex"],
    },
    {
        "column_name": "opex_fix_costs_total",
        "operation": compute_opex_fix_costs,
        "description": "Operation and maintenance costs are calculated by multiplying the optimized capacity "
        "by the OPEX",
        "argument_names": ["aggregated_flow", "opex_fix"],
    },
    {
        "column_name": "variable_costs_total",
        "operation": compute_variable_costs,
        "description": "Variable costs are calculated by multiplying the total flow "
        "by the marginal/carrier costs",
        "argument_names": ["aggregated_flow", "marginal_cost", "carrier_cost"],
    },
    {
        "column_name": "renewable_generation",
        "operation": compute_renewable_generation,
        "description": "The renewable generation for each component is computed from the flow and the "
        "renewable factor.",
        "argument_names": [
            "aggregated_flow",
            "renewable_factor",
        ],
    },
    {
        "column_name": "co2_emissions",
        "operation": compute_co2_emissions,
        "description": "CO2 emissions are calculated from the flow and the emission factor",
        "argument_names": ["aggregated_flow", "emission_factor"],
    },
    {
        "column_name": "land_requirement_additional",
        "operation": compute_land_requirement_additional,
        "description": "The additional land requirement calculates the land required for the optimized capacities",
        "argument_names": ["investments", "land_requirement_factor"],
    },
    {
        "column_name": "land_requirement_total",
        "operation": compute_land_requirement_total,
        "description": "The total land requirement calculates the land required for the total capacities",
        "argument_names": ["capacity_total", "land_requirement_factor"],
    },
    {
        "column_name": "water_footprint",
        "operation": compute_water_footprint,
        "description": "The water footprint calculates the water footprint for the aggregated flows of each component",
        "argument_names": ["aggregated_flow", "water_footprint_factor"],
    },
]

# ToDo: turn dict into a class (see CALCULATED_OUTPUTS) and decide where this belongs - either here or in processing
#  or maybe these can be joined with CALCULATED_OUTPUTS and there is another parameter that defines if it is a calculation
#  per component (to be added to df_results) or a calculation for the whole system (e.g. LCOE, total emissions etc).
#  Probably this should be included with the other CALCULATED_OUTPUTS eventually, but should ask PF
CALCULATED_KPIS = [
    {
        "column_name": "annuity_total",
        "operation": compute_system_annuity_total,
        "description": "The system total annuity is calculated by summing up the total annuity for each component",
        "argument_names": ["annuity_total"],
    },
    {
        "column_name": "variable_costs_total",
        "operation": compute_system_variable_costs_total,
        "description": "The system total variable costs is calculated by summing up the total variable costs for "
        "each component flow",
        "argument_names": ["variable_costs_total"],
    },
    {
        "column_name": "system_cost_total",
        "operation": compute_system_cost_total,
        "description": "The total system cost is calculated by adding the total annuity to the total variable costs",
        "argument_names": ["annuity_total", "variable_costs_total"],
    },
    {
        "column_name": "total_upfront_investments",
        "operation": compute_system_upfront_investments_total,
        "description": "The total upfront investments value is calculated by summing the upfront investment"
        "costs for each component",
        "argument_names": ["upfront_investment_costs"],
    },
    {
        "column_name": "specific_system_cost",
        "operation": compute_specific_system_cost,
        "description": "T",
        "argument_names": ["aggregated_flow", "total_annuity", "total_variable_costs"],
    },
    {
        "column_name": "co2_emissions_total",
        "operation": compute_system_co2_emissions_total,
        "description": "The total emissions is calculated by summing the c02 emissions "
        "for each component",
        "argument_names": ["co2_emissions"],
    },
    {
        "column_name": "land_requirement_additional",
        "operation": compute_system_land_requirement_additional,
        "description": "The total additional land requirement is calculated by summing the additional land requirement "
        "for each component",
        "argument_names": ["land_requirement_additional"],
    },
    {
        "column_name": "land_requirement_total",
        "operation": compute_system_land_requirement_total,
        "description": "The total land requirement is calculated by summing the total land requirement "
        "for each component",
        "argument_names": ["land_requirement_total"],
    },
    {
        "column_name": "total_water_footprint",
        "operation": compute_water_footprint_total,
        "description": "The total water footprint is calculated by summing the water footprint required "
        "for each component",
        "argument_names": ["water_footprint"],
    },
    {
        "column_name": "renewable_share",
        "operation": compute_renewable_share,
        "description": "The renewable share is calculated by dividing the renewable generation by the total "
        "generation",
        "argument_names": ["renewable_factor", "aggregated_flow"],
    },
]

# Add docstrings from function handles for documentation purposes
for calc in CALCULATED_OUTPUTS + CALCULATED_KPIS:
    func_handle = calc.get("operation", None)
    if callable(func_handle):
        calc["docstring"] = func_handle.__doc__
    else:
        calc["docstring"] = ""


def _validate_calculation(calculation):
    """Check if the parameters of a calculation are there and of the right format"""
    var_name = calculation.get("column_name", None)
    fhandle = calculation.get("operation", None)

    if var_name is None:
        raise ValueError(
            f"The 'column_name' under which the calculation should be saved in the results DataFrame is missing from the calculation dict: {calc}. Please check your input or look at help(apply_calculations) for the formatting of the calculation dict"
        )

    if not callable(fhandle):
        raise ValueError(
            f"The provided function handle for calculation of column '{var_name}' is not callable"
        )


def infer_busses_carrier(energy_system):
    """Loop through the nodes of an energy system and infer the carrier of busses from them

    Parameters
    ----------
    energy_system: oemof.solph.EnergySystem instance

    Returns
    -------
    dict mapping the busses labels to their carrier

    """

    busses_carrier = {}

    for node in energy_system.nodes:
        if hasattr(node, "carrier"):
            # quick fix to work for MIMO component
            # ToDo: assign carrier to busses instead of components to avoid problems
            for attribute in ("bus", "from_bus", "from_bus_0", "to_bus_1"):
                if hasattr(node, attribute):

                    bus_label = getattr(node, attribute).label
                    if bus_label in busses_carrier:
                        if busses_carrier[bus_label] != node.carrier:
                            raise ValueError(
                                f"Two different carriers ({busses_carrier[bus_label]}, {node.carrier}) are associated to the same bus '{bus_label}'"
                            )
                    else:
                        busses_carrier[bus_label] = node.carrier

    busses = [node.label for node in energy_system.nodes if isinstance(node, solph.Bus)]

    for bus_label in busses:
        if bus_label not in busses_carrier:
            raise ValueError(
                f"Bus '{bus_label}' is missing from the busses carrier dict inferred from the EnergySystem instance"
            )

    return busses_carrier


def infer_asset_types(energy_system):
    """Loop through the nodes of an energy system and infer their types

    Parameters
    ----------
    energy_system: oemof.solph.EnergySystem instance

    Returns
    -------
    a dict mapping the asset (nodes which are not busses) labels to their type

    """
    asset_types = {}
    for node in energy_system.nodes:
        if isinstance(node, solph.Bus) is False:
            asset_types[node.label] = node.type
    return asset_types


def construct_multi_index_levels(flow_tuple, busses_info, assets_info=None):
    """Infer the index levels of the multi index dataframe sequence tuple and extra optional information

    Parameters
    ----------
    flow_tuple: tuple of bus label and asset label
        (A,B) means flow from A to B
    busses_info: either a list or a dict
        if not a list of busses labels, then a dict where busses labels are keys mapped to the bus carrier
    assets_info: dict
        mapping of asset labels to their type

    Returns
    -------
    a tuple with (bus label, direction of flow relative to asset, asset label, bus carrier (optional), asset type (optional)) direction is either 'in' or 'out'.

    The minimal tuple (b_elec, "in", demand) would read the flow goes from bus 'b_elec' '"in"' asset 'demand'

    """
    if isinstance(busses_info, dict):
        busses_labels = [bn for bn in busses_info]
    elif isinstance(busses_info, list):
        busses_labels = busses_info

    # infer which of the 2 nodes composing the flow is the bus
    bus_label = set(busses_labels).intersection(set(flow_tuple))
    if len(bus_label) == 1:
        bus_label = bus_label.pop()
    else:
        raise ValueError("Flow tuple does not contain only one bus node")
    # get position of bus node in the flow tuple
    idx_bus = flow_tuple.index(bus_label)
    answer = None

    # determine whether the flow goes from bus to asset or reverse
    if idx_bus == 0:
        # going from bus to asset, so the flow goes in to the asset
        asset_label = flow_tuple[1]
        answer = (bus_label, "in", asset_label)

    elif idx_bus == 1:
        asset_label = flow_tuple[0]
        # going from asset to bus, so the flow goes out of the asset
        answer = (bus_label, "out", asset_label)

    # add information of the bus carrier, if provided
    if isinstance(busses_info, dict):
        answer = answer + (busses_info[bus_label],)
    # add information of the asset type, if provided
    if assets_info is not None:
        answer = answer + (assets_info[asset_label],)
    return answer


def construct_dataframe_from_results(energy_system, bus_carrier=True, asset_type=True):
    """

    Parameters
    ----------
    energy_system: oemof.solph.EnergySystem instance
    bus_carrier: bool (opt)
        If set to true, the multi-index of the DataFrame will have a level about bus carrier
    asset_type: bool (opt)
        If set to true, the multi-index of the DataFrame will have a level about the asset type


    Returns
    -------
    Dataframe with oemof result sequences's timestamps as columns as well as investment and a multi-index built automatically, see construct_multi_index_levels for more information on the multi-index
    """
    mi_levels = [
        "bus",
        "direction",
        "asset",
    ]

    busses_info = infer_busses_carrier(energy_system)
    if bus_carrier is False:
        busses_info = list(busses_info.keys())
    else:
        mi_levels.append("carrier")

    if asset_type is True:
        assets_info = infer_asset_types(energy_system)
        mi_levels.append("facade_type")
    else:
        assets_info = None

    results = energy_system.results

    if isinstance(results, dict):
        ts = []
        investments = []
        flows = []
        for x, res in solph.views.convert_keys_to_strings(results).items():
            # filter out entries where the second element of the tuple is 'None' and ensure the
            # tuple has exactly two elements
            if x[1] != "None" and len(x) == 2:
                col_name = res["sequences"].columns[0]
                ts.append(
                    res["sequences"].rename(
                        columns={col_name: x, "variable_name": "timesteps"}
                    )
                )
                # here change this information for flow_tuple in ('mimo', 'in_group_0', '0')
                flows.append(
                    construct_multi_index_levels(
                        x, busses_info=busses_info, assets_info=assets_info
                    )
                )

                invest = None if res["scalars"].empty is True else res["scalars"].invest
                investments.append(invest)
        ts_df = pd.concat(ts, axis=1, join="inner")
        mindex = pd.MultiIndex.from_tuples(flows, names=mi_levels)

        df = pd.DataFrame(
            data=ts_df.T.to_dict(orient="split")["data"],
            index=mindex,
            columns=ts_df.index,
        )

        df["investments"] = investments
        df.sort_index(inplace=True)

    return df


def process_raw_results(df_results):
    """Compute the min, max and aggregated flows for each asset-bus pair

    Parameters
    ----------
    df_results: pandas DataFrame
        the outcome of construct_dataframe_from_results()

    Returns
    -------
    """
    temp = df_results[df_results.columns.difference(RAW_OUTPUTS)]
    df_results["flow_min"] = temp.min(axis=1)
    df_results["flow_max"] = temp.max(axis=1)
    df_results["aggregated_flow"] = temp.sum(axis=1)
    return df_results


def process_raw_inputs(df_results, dp_path, raw_inputs=RAW_INPUTS, typemap=None):
    """Find the input parameters from the datapackage.json file


    Parameters
    ----------
    df_results: pandas DataFrame
        the outcome of construct_dataframe_from_results()
    dp_path: string
        path to the datapackage.json file
    raw_inputs: list of string
        list of parameters from the datapackage one would like to collect for result post-processing

    Returns
    -------

    """
    if typemap is None:
        typemap = {}

    p = Package(dp_path)
    # initialise inputs_df with raw inputs as indexes
    inputs_df = pd.DataFrame(index=raw_inputs)
    # inputs_df = None
    for r in p.resources:
        if "elements" in r.descriptor["path"] and r.name != "bus":
            df = pd.DataFrame.from_records(r.read(keyed=True), index="name")
            resource_inputs = df[list(set(raw_inputs).intersection(set(df.columns)))].T
            if inputs_df is None:
                if not resource_inputs.empty:
                    inputs_df = resource_inputs
            else:
                inputs_df = inputs_df.join(resource_inputs)
            # if r.name in typemap:
            # TODO here test if facade_type has the method 'validate_datapackage'
            #   inputs_df = typemap[r.name].processing_raw_inputs(r, inputs_df)

    # kick out the lines where all values are NaN
    inputs_df = inputs_df.dropna(how="all")
    # append the inputs of the datapackage to the results DataFrame
    inputs_df.T.index.name = "asset"
    return df_results.join(inputs_df.T.apply(pd.to_numeric, downcast="float"))


def apply_calculations(results_df, calculations=CALCULATED_OUTPUTS):
    """Apply calculation and populate the columns of the results_df

    Parameters
    ----------
    df_results: pandas DataFrame
        the outcome of process_raw_input()
    calculations: list of dict
        each dict should contain
            "column_name" (the name of the new column within results_df),
            "operation" (handle of a function which will be applied row-wise to results_df),
            "description" (a string for documentation purposes)
            and "argument_names" (list of columns needed within results_df)

    Returns
    -------

    """
    for calc in calculations:
        _validate_calculation(calc)
        var_name = calc.get("column_name")
        argument_names = calc.get("argument_names", [])
        func_handle = calc.get("operation")
        try:
            _check_arguments(results_df, column_names=argument_names, col_name=var_name)
        except AttributeError as e:
            logging.warning(e)
            continue

        results_df[var_name] = results_df.apply(
            func_handle,
            axis=1,
        )
        # ToDo: I've commented this out for now but decide if this or some form should be kept in
        # # check if the new column contains all None values and remove it if so
        # if results_df[var_name].isna().all():
        #     results_df.drop(columns=[var_name], inplace=True)
        #     logging.info(
        #         f"Removed column '{var_name}' because it contains all None values."
        #     )


def apply_kpi_calculations(results_df, calculations=CALCULATED_KPIS):
    """Apply calculation and return a new DataFrame with the KPIs.

    Parameters
    ----------
    results_df : pd.DataFrame
        The input DataFrame with raw data.
    calculations : list of dict
        List of calculations to be applied. Each calculation is a dictionary
        with keys: "column_name", "argument_names", and "operation".

    Returns
    -------
    pd.DataFrame
        A new DataFrame containing the calculated KPI values with var_name as the index.
    """
    kpis = []

    for calc in calculations:
        _validate_calculation(calc)
        var_name = calc.get("column_name")
        argument_names = calc.get("argument_names", [])
        func_handle = calc.get("operation")

        try:
            _check_arguments(results_df, column_names=argument_names, col_name=var_name)
        except AttributeError as e:
            logging.warning(e)
            continue

        kpi_value = func_handle(results_df)
        kpis.append({"kpi": var_name, "value": kpi_value})

    kpi_df = pd.DataFrame(kpis).set_index("kpi")
    return kpi_df



from dataclasses import field
from typing import Sequence, Union

from oemof.solph._plumbing import sequence
from oemof.solph.buses import Bus
from oemof.solph.components import Converter
from oemof.solph.flows import Flow

from oemof.tabular._facade import Facade, dataclass_facade


@dataclass_facade
class PVPanel(Converter, Facade):
    r"""PV panel unit with one input and one output. The temperature factor
    is calculated and considered within the electricity generation.

    Parameters
    ----------
    from_bus: oemof.solph.Bus
        An oemof bus instance where the PV panel unit is connected to with
        its input.
    to_bus: oemof.solph.Bus
        An oemof bus instance where the PV panel unit is connected to with
        its output.
    capacity: numeric
        The conversion capacity (output side) of the unit.
    marginal_cost: numeric
        Marginal cost for one unit of produced output. Default: 0
    carrier_cost: numeric
        Carrier cost for one unit of used input. Default: 0
    capacity_cost: numeric
        Investment costs per unit of output capacity.
        If capacity is not set, this value will be used for optimizing the
        conversion output capacity.
    expandable: boolean or numeric (binary)
        True, if capacity can be expanded within optimization. Default: False.
    capacity_potential: numeric
        Maximum invest capacity in unit of output capacity.
    capacity_minimum: numeric
        Minimum invest capacity in unit of output capacity.
    input_parameters: dict (optional)
        Set parameters on the input edge of the conversion unit
        (see oemof.solph for more information on possible parameters)
    output_parameters: dict (optional)
        Set parameters on the output edge of the conversion unit
         (see oemof.solph for more information on possible parameters)
    t_air: array-like
        Ambient air temperature
    ghi: array-like
        Global horizontal irradiance
    p_rpv: numeric
        Rated power of photovoltaic panel. Default: 270
    r_ref: numeric
        Solar radiation at reference conditions. Default: 1000
    n_t: numeric
        Temperature coefficient of PV panel. Default: -0.0037
    t_c_ref: numeric
        Cell temperature at reference conditions. Default: 25
    noct: numeric
        Normal operating cell temperature. Default: 48

    SHOULD INCLUDE FUNCTIONS AND EXAMPLE HERE

    """

    from_bus: Bus

    to_bus: Bus

    carrier: str

    tech: str

    t_air: Union[float, Sequence[float]]

    ghi: Union[float, Sequence[float]]

    capacity: float = None

    marginal_cost: float = 0

    carrier_cost: float = 0

    capacity_cost: float = None

    expandable: bool = False

    capacity_potential: float = float("+inf")

    capacity_minimum: float = None

    input_parameters: dict = field(default_factory=dict)

    output_parameters: dict = field(default_factory=dict)

    p_rpv: float = 270

    r_ref: float = 1000

    n_t: float = -0.0037

    t_c_ref: float = 25

    noct: float = 48

    def build_solph_components(self):
        """ """
        if self.t_air is None or self.ghi is None:
            # handle the case when t_air or ghi is None
            print("Error: t_air or ghi is None. Cannot perform calculations.")
            return
        # assign the air temperature and solar irradiance
        t_air_values = self.t_air
        ghi_values = self.ghi
        # raise error if air temperature list and solar irradiance list are different lengths
        if len(t_air_values) != len(ghi_values):
            raise ValueError("Length mismatch between t_air and ghi profiles.")
        # calculates the temperature factor values
        pv_tf_values = []
        for t_air, ghi in zip(t_air_values, ghi_values):
            t_c = t_air + ((self.noct - 20) / 800) * ghi
            pv_tf = (
                self.p_rpv * (1 / self.r_ref) * (1 + self.n_t * (t_c - self.t_c_ref))
            )
            pv_tf_values.append(pv_tf)

        self.conversion_factors.update(
            {
                self.from_bus: sequence(1),
                self.to_bus: sequence(pv_tf_values),
            }
        )

        self.inputs.update(
            {
                self.from_bus: Flow(
                    variable_costs=self.carrier_cost, **self.input_parameters
                )
            }
        )

        self.outputs.update(
            {
                self.to_bus: Flow(
                    nominal_value=self._nominal_value(),
                    variable_costs=self.marginal_cost,
                    investment=self._investment(),
                    **self.output_parameters,
                )
            }
        )

        def processing_raw_inputs(self, resource, results_df):
            # function to apply on df from above

            return results_df

        def validate_datapackage(self, resource):
            # modify the resource (datapackage.resource)
            # should it return the resource?
            pass



import logging
import os
import warnings

import pandas as pd
from oemof.tabular.postprocessing import calculations as clc, naming
from oemof.tabular.postprocessing.core import Calculator

from oemof_tabular_plugins.datapackage.post_processing import (
    construct_dataframe_from_results,
    process_raw_results,
    process_raw_inputs,
    apply_calculations,
    apply_kpi_calculations,
)

# ------ New post-processing to create tables ------
# This dictionary contains groups of columns that should be extracted from the df_results to generate a clearer overview


# TODO add a column for planned capacity (not optimizable but including costs) in capacities if it gets properly
#  implemented (planned capacity can be set by setting capacity_minimum == capacity_potential and dispatchable = True
RESULT_TABLE_COLUMNS = {
    "costs": ["upfront_investment_costs", "total_annuity", "total_variable_costs"],
    "capacities": [
        "capacity",
        "storage_capacity",
        "capacity_minimum",
        "capacity_potential",
        "storage_capacity_potential",
        "expandable",
        "investments",
        "total_capacity",
    ],
}


def extract_table_from_results(df_results, columns):
    """Extracts a set of columns from the df_results DataFrame. The lists of columns to generate these tables can be
    defined in RESULT_TABLE_COLUMNS
    :param df_results: multiindex results dataframe with additional columns (OTPCalculator.df_results)
    :param columns: list of columns to generate the sub-table (should be defined in RESULT_TABLE_COLUMNS)
    :return: dataframe containing the columns specified in columns, if present in df_results
    """
    missing_columns = []
    for col in columns:
        try:
            df_results[col]
        except KeyError:
            # If the key is not in the df_results dataframe, log it as a warning and pop the column from the list
            logging.warning(
                f"The column {col} was not found in the results DataFrame, will be skipped in the subtable"
            )
            missing_columns.append(col)

    columns = [col for col in columns if col not in missing_columns]
    results_table = df_results[columns].copy()
    # TODO some of these names may be confusing because they are just the columns, maybe there should be a
    #  verbose parameter in CALCULATED OUTPUTS that we can then also use here
    results_table.columns = [
        col.title().replace("_", " ") for col in results_table.columns
    ]
    return results_table


def save_table_to_csv(table, results_path, filename):
    """Saves a DataFrame to a .csv file"""
    filepath = os.path.join(results_path, filename)
    table.to_csv(filepath)


# TODO figure out the best table/display for storage results


# --------------------------------------------------
class OTPCalculator(Calculator):
    def __init__(self, input_parameters, energy_system, dp_path):

        self.df_results = construct_dataframe_from_results(energy_system)
        self.df_results = process_raw_results(self.df_results)
        self.df_results = process_raw_inputs(self.df_results, dp_path)
        apply_calculations(self.df_results)
        self.kpis = apply_kpi_calculations(self.df_results)

        super().__init__(input_parameters, energy_system.results)


def post_processing(params, es, results_path, dp_path):
    # ToDo: adapt this function after multi-index dataframe is implemented to make it more concise / cleaner
    # ToDo: params can be accessed in results so will not need to be a separate argument
    """
    The main post-processing function extracts various scalar and timeseries data and stores it in CSV files.
    :param params: energy system parameters
    :param es: oemof energy_system with results in it, ie es.results = processing.results(m) has been performed
    :param results_path: results directory path
    :param dp_path: path to the datapackage.json file
    """
    # initiate calculator for post-processing
    calculator = OTPCalculator(params, es, dp_path)
    # print(calculator.df_results)
    results = es.results
    results_by_flow = calculator.df_results
    results_by_flow.to_csv(results_path + "/all_results_by_flow.csv", index=True)
    kpis = calculator.kpis
    kpis.to_csv(results_path + "/kpis.csv", index=True)

    # get sub-tables from results dataframe
    cost_table = extract_table_from_results(
        calculator.df_results, RESULT_TABLE_COLUMNS["costs"]
    )
    capacities_table = extract_table_from_results(
        calculator.df_results, RESULT_TABLE_COLUMNS["capacities"]
    )

    # save tables to csv files
    tables_to_save = {"costs.csv": cost_table, "capacities.csv": capacities_table}
    for filename, table in tables_to_save.items():
        save_table_to_csv(table, results_path, filename)

    # ----- OLD POST-PROCESSING - TO BE DELETED ONCE CERTAIN -----

    # calculate scalars using functions from clc module
    aggregated_flows = clc.AggregatedFlows(calculator).result
    storage_losses = clc.StorageLosses(calculator).result
    transmission_losses = clc.TransmissionLosses(calculator).result
    invested_capacity = clc.InvestedCapacity(calculator).result
    invested_storage_capacity = clc.InvestedStorageCapacity(calculator).result
    invested_capacity_costs = clc.InvestedCapacityCosts(calculator).result
    invested_storage_capacity_costs = clc.InvestedStorageCapacityCosts(
        calculator
    ).result
    summed_carrier_costs = clc.SummedCarrierCosts(calculator).result
    summed_marginal_costs = clc.SummedMarginalCosts(calculator).result
    total_system_costs = clc.TotalSystemCosts(calculator).result

    # combine all results into a single dataframe
    all_scalars = [
        aggregated_flows,
        storage_losses,
        transmission_losses,
        invested_capacity,
        invested_storage_capacity,
        invested_capacity_costs,
        invested_storage_capacity_costs,
        summed_carrier_costs,
        summed_marginal_costs,
    ]
    # map variable names and add component information
    all_scalars = pd.concat(all_scalars, axis=0)
    all_scalars = naming.map_var_names(
        all_scalars,
        calculator.scalar_params,
        calculator.busses,
        calculator.links,
    )
    all_scalars = naming.add_component_info(all_scalars, calculator.scalar_params)
    print("Total System Cost", total_system_costs)
    total_system_costs.index.names = ("name", "var_name")
    all_scalars = pd.concat([all_scalars, total_system_costs], axis=0)
    all_scalars = all_scalars.sort_values(by=["carrier", "tech", "var_name"])
    # save all scalar results to a csv file
    filepath_name_all_scalars = os.path.join(results_path, "all_scalars.csv")
    all_scalars.to_csv(filepath_name_all_scalars)

    # saves all hourly timeseries as a dataframe (see test_postprocessing.py in oemof-tabular/tests
    # for example if wanting to filter particular nodes)
    all_sequences = clc.AggregatedFlows(calculator, resample_mode="H")
    # save all timeseries (sequence) results to a csv file
    filepath_name_all_sequences = os.path.join(results_path, "all_sequences.csv")
    all_sequences.sequences.to_csv(filepath_name_all_sequences)

    return



"""Only meant to be used for `pip install -e .`"""

from setuptools import setup, find_packages, Extension

setup_args = dict(
    packages=find_packages(where="src"), package_dir={"": "src"}  # list  # mapping
)

setup(**setup_args)
